/*
** pbm.c -- PBM file handling routines for ShowTime
**
** Copyright (C) 1994 by Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
*/

#include "showtime.h"
#include "pbm.h"
#include "cmap.h"
#include "video.h"
#include "file.h"

/* ====================================================================== */

/* sc_getc: get a character, skipping comments */
static char
sc_getc(FILE *fio)
{
  char ch;

  GET_BYTE(fio, ch);

  if(ch=='#') {
    do {
      GET_BYTE(fio, ch);
    } while (ch != '\n' && ch != '\r');
  }
  return(ch);
}

/* ---------------------------------------------------------------------- */

/* sc_getint : get from fio a text integer, skipping comments */
int
sc_getint(FILE *fio)
{
  register char ch;
  register int i;
  char ci[10];

  i=0;
  do {
    ch=sc_getc(fio);
    i++;
  } while (ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r');

  if(!isdigit(ch)) {
    logmesg("sc_getint: junk %i=%c after %i white spaces, should be integer\n",
            ch, ch, i);
    return(EOF);
  }

  i=0;
  do {
    ci[i]=ch;
    i++;
    ch=sc_getc(fio);
  } while (isdigit(ch));
  ci[i]='\0';
  i=atoi(ci);
  return(i);
}

/* ---------------------------------------------------------------------- */

/* taken directly from the PBM library */
static unsigned char
pbm_getbit(FILE *fio)
{
  register char ch;

  do {
    ch = sc_getc( fio );
  } while ( ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r' );

  if ( ch != '0' && ch != '1' )
    logmesg("pbm_getbit: junk in file where bits should be\n");

  return ( ch == '1' ) ? 1 : 0;
}

/* ---------------------------------------------------------------------- */

/* load_pnm_header : read a PNM image header and return pertinent info
** 7apr94 mjg
*/
int
load_pnm_header(image_hdr_t *ihdrP, screendata_t *sdP, FILE *fio)
{
  pbm_hdr_t *pbmP;

  if((pbmP = ihdrP->hdr = malloc(sizeof(pbm_hdr_t))) == NULL) {
    logmesg("load_pnm_header: out of mem allocating pbmP\n");
    exit(1);
  }

  pbmP->format = ihdrP->image_type;

  switch(pbmP->format) {
    case PBM_FORMAT: /* ASCII PBM */
    case RPBM_FORMAT: /* Raw PBM */
      ihdrP->load_img = load_pbm_image;

      /* Read the width and height */
      NONEGA(ihdrP->num_cols, "rih: PBM num_cols");
      NONEGA(ihdrP->num_rows, "rih: PBM num_rows");
      break;

    case RPGM_FORMAT: /* Raw PGM */
      ihdrP->load_img = load_pgm_image;

      /* Read the width and height */
      NONEGA(ihdrP->num_cols, "rih: PBM num_cols");
      NONEGA(ihdrP->num_rows, "rih: PBM num_rows");

      sdP->xoffset = 0;
      sdP->xoffset = 0;

      /* Read the maxval */
      NONEGA(pbmP->maxval, "rih: PBM maxval");
      break;

    case PPM_FORMAT: /* ASCII PPM */
    case RPPM_FORMAT: /* Raw PPM */
      ihdrP->load_img = load_ppm_image;

      /* Read the width and height */
      NONEGA(ihdrP->num_cols, "rih: PBM num_cols");
      NONEGA(ihdrP->num_rows, "rih: PBM num_rows");

      sdP->xoffset = 0;
      sdP->xoffset = 0;

      /* Read the maxval */
      NONEGA(pbmP->maxval, "rih: PBM maxval");

      /* Set the full-color approximation method */
      pbmP->color_approx=RGB332;

      break;

    default:
      logmesg("load_pnm_header: bad PNM format %i\n", pbmP->format);
      return(1);
      break;
  }

  return(0);
}

/* ---------------------------------------------------------------------- */

/* load_ppm_palette: read a PPM file and use it as a palette.
** also install the palette
*/
int
load_ppm_palette(FILE *fio, image_hdr_t *ihdrP)
{
  register int  i;
  int           num_vals;
  int           red, green, blue;
  unsigned char rgb_map[256][3];
  pbm_hdr_t     *pbmP = ihdrP->hdr;

  num_vals = ihdrP->num_cols * ihdrP->num_rows;
  if(num_vals > 256) {
    logmesg("load_ppm_palette: too many palette colors!  Using first 256.\n");
    num_vals = 256;
  }

  /* Read each palette color value and install it */
  for(i=0; i<num_vals; i++) {
    /* Read each red green and blue component */
    switch(pbmP->format) {
      case PPM_FORMAT:
        red   = sc_getint(fio);
        green = sc_getint(fio);
        blue  = sc_getint(fio);
        break;
      case RPPM_FORMAT:
        GET_BYTE(fio, red);
        GET_BYTE(fio, green);
        GET_BYTE(fio, blue);
        break;
      case PGM_FORMAT:
        red=green=blue=sc_getint(fio);
        break;
      case RPGM_FORMAT:
        GET_BYTE(fio, red);
        green=blue=red;
        break;
    }
    if(red>pbmP->maxval)
      logmesg("load_ppm_pal:red[%i]=%i>maxval (%i)\n", i,red,pbmP->maxval);

    if(green>pbmP->maxval)
      logmesg("load_ppm_pal:green[%i]=%i>maxval (%i)\n", i,green,pbmP->maxval);

    if(blue>pbmP->maxval)
      logmesg("load_ppm_pal:blue[%i]=%i>maxval (%i)\n", i,blue,pbmP->maxval);

    /* Scale the color values to be from 0 to 255 */
    rgb_map[i][0] = red   = (int)(0.5+(double)red   * 255 / pbmP->maxval);
    rgb_map[i][1] = green = (int)(0.5+(double)green * 255 / pbmP->maxval);
    rgb_map[i][2] = blue  = (int)(0.5+(double)blue  * 255 / pbmP->maxval);
  }

  /* Install the colormap */
  install_cmap(num_vals, rgb_map, 255);

  return(0);
}

/* ---------------------------------------------------------------------- */

/* load_pbm_image: read a PBM image
*/
int
load_pbm_image(image_hdr_t *ihdrP, screendata_t *sdP, FILE *fio)
{
  register int           col;
  register int           row;
  register int           bitshift;
  register unsigned char byte;
  pbm_hdr_t              *pbmP = ihdrP->hdr;

  /* Load each portion of the image */
  for(row=0; row<sdP->num_rows; row++) {
    switch(pbmP->format) {
      case PBM_FORMAT:
        for(col=0; col<sdP->num_cols; col++) {
          *(sdP->data[row]+sdP->header_size+col) = pbm_getbit(fio);
        }
        break;

      case RPBM_FORMAT:
        for(bitshift= -1, col=0; col<sdP->num_cols; col++) {
          if(bitshift == -1) {
            GET_BYTE(fio, byte);
            bitshift=7;
          }
          *(sdP->data[row]+sdP->header_size+col) = (byte >> bitshift) & 1;
          --bitshift;
        }
        break;

      default:
        logmesg("load_pbm_image: bad format\n");
        return(EOF);
    }
  }
  return(0);
}

/* ---------------------------------------------------------------------- */

/* load_pgm_image: read a PGM image
*/
int
load_pgm_image(image_hdr_t *ihdrP, screendata_t *sdP, FILE *fio)
{
  register int    row;

  /* Load each portion of the image */
  for(row=0; row<sdP->num_rows; row++) {
    /* Read the image data */
    if(get_block(fio, sdP->data[row]+sdP->header_size, sdP->num_cols)) {
      logmesg("\nload_pgm_image: unexpected read error in row %i\n", row);
      return(EOF);
    }
  }
  return(0);
}

/* ---------------------------------------------------------------------- */

#define RM (7<<5)
#define GM (7<<2)
#define BM (3)

/* load_ppm_image: read a PPM image
** 25mar94 MJG
*/
int
load_ppm_image(image_hdr_t *ihdrP, screendata_t *sdP, FILE *fio)
{
  register int row, col;
  register int ri, gi, bi;
  register int rb, gb, bb;
  register int rerr=0, gerr=0, berr=0;
  register int pixel;
  unsigned char *img_row;
  pbm_hdr_t              *pbmP = ihdrP->hdr;
  unsigned char rgb332[256][3];

  /* Load palette table to compare truncation error */
  if(pbmP->color_approx == RGB332) {
    make_palette(RGB332, rgb332);
  }

  /* Load each portion of the image */
  for(row=0; row<sdP->num_rows; row++) {
    /* Read the image data */
    switch(pbmP->format) {
      case PPM_FORMAT:
        img_row = sdP->data[row]+sdP->header_size;
        for(col=0; col<sdP->num_cols; col++) {
          ri=sc_getint(fio);
          gi=sc_getint(fio);
          bi=sc_getint(fio);
          if(pbmP->color_approx==GRAY) {
            pixel = .229 * ri + .587 * gi + .114 * bi;
          } else if (pbmP->color_approx==RGB332) {
            RGB332MAP(pixel);
          } else {
            logmesg("load_ppm_image: invalid approximation method\n");
            return(1);
          }
          *(img_row + col) = pixel;
        }
        break;

      case RPPM_FORMAT:
        img_row = sdP->data[row]+sdP->header_size;
        for(col=0; col<sdP->num_cols; col++) {
          GET_BYTE(fio, ri);
          GET_BYTE(fio, gi);
          GET_BYTE(fio, bi);
          if(pbmP->color_approx==GRAY) {
            pixel = .229 * ri + .587 * gi + .114 * bi;
          } else if (pbmP->color_approx==RGB332) {
            RGB332MAP(pixel);
          } else {
            logmesg("load_ppm_image: invalid approximation method\n");
            return(1);
          }
          *(img_row + col) = pixel;
        }
        break;

      default:
        break;
    }
  }
  return(0);
}
